from __future__ import division
import numpy as np

from .elemwise import elemwise1, elemwise2, ielemwise2, compare, arg, GpuElemwise, as_argument
from .reduction import reduce1
from .dtypes import dtype_to_ctype, get_np_obj, get_common_dtype
from . import gpuarray


class ndgpuarray(gpuarray.GpuArray):
    """
    Extension class for gpuarray.GpuArray to add numpy mathematical
    operations between arrays.  These operations are all performed on
    the GPU but this is not the most efficient way since it will
    involve the creation of temporaries (just like numpy) for all
    intermediate results.

    This class may help transition code from numpy to pygpu by acting
    more like a drop-in replacement for numpy.ndarray than the raw
    GpuArray class.
    """
    # add
    def __add__(self, other):
        return elemwise2(self, '+', other, self, broadcast=True)

    def __radd__(self, other):
        return elemwise2(other, '+', self, self, broadcast=True)

    def __iadd__(self, other):
        return ielemwise2(self, '+', other, broadcast=True)

    # sub
    def __sub__(self, other):
        return elemwise2(self, '-', other, self, broadcast=True)

    def __rsub__(self, other):
        return elemwise2(other, '-', self, self, broadcast=True)

    def __isub__(self, other):
        return ielemwise2(self, '-', other, broadcast=True)

    # mul
    def __mul__(self, other):
        return elemwise2(self, '*', other, self, broadcast=True)

    def __rmul__(self, other):
        return elemwise2(other, '*', self, self, broadcast=True)

    def __imul__(self, other):
        return ielemwise2(self, '*', other, broadcast=True)

    # div
    def __div__(self, other):
        return elemwise2(self, '/', other, self, broadcast=True)

    def __rdiv__(self, other):
        return elemwise2(other, '/', self, self, broadcast=True)

    def __idiv__(self, other):
        return ielemwise2(self, '/', other, broadcast=True)

    # truediv
    def __truediv__(self, other):
        np1 = get_np_obj(self)
        np2 = get_np_obj(other)
        res = (np1.__truediv__(np2)).dtype
        return elemwise2(self, '/', other, self, odtype=res, broadcast=True)

    def __rtruediv__(self, other):
        np1 = get_np_obj(self)
        np2 = get_np_obj(other)
        res = (np2.__truediv__(np1)).dtype
        return elemwise2(other, '/', self, self, odtype=res, broadcast=True)

    def __itruediv__(self, other):
        np2 = get_np_obj(other)
        kw = {'broadcast': True}
        if self.dtype == np.float32 or np2.dtype == np.float32:
            kw['op_tmpl'] = "a = (float)a / (float)b"
        if self.dtype == np.float64 or np2.dtype == np.float64:
            kw['op_tmpl'] = "a = (double)a / (double)b"
        return ielemwise2(self, '/', other, **kw)

    # floordiv
    def __floordiv__(self, other):
        out_dtype = get_common_dtype(self, other, True)
        kw = {'broadcast': True}
        if out_dtype.kind == 'f':
            kw['op_tmpl'] = "res = floor((%(out_t)s)a / (%(out_t)s)b)"
        return elemwise2(self, '/', other, self, odtype=out_dtype, **kw)

    def __rfloordiv__(self, other):
        out_dtype = get_common_dtype(other, self, True)
        kw = {'broadcast': True}
        if out_dtype.kind == 'f':
            kw['op_tmpl'] = "res = floor((%(out_t)s)a / (%(out_t)s)b)"
        return elemwise2(other, '/', self, self, odtype=out_dtype, **kw)

    def __ifloordiv__(self, other):
        out_dtype = self.dtype
        kw = {'broadcast': True}
        if out_dtype == np.float32:
            kw['op_tmpl'] = "a = floor((float)a / (float)b)"
        if out_dtype == np.float64:
            kw['op_tmpl'] = "a = floor((double)a / (double)b)"
        return ielemwise2(self, '/', other, **kw)

    # mod
    def __mod__(self, other):
        out_dtype = get_common_dtype(self, other, True)
        kw = {'broadcast': True}
        if out_dtype.kind == 'f':
            kw['op_tmpl'] = "res = fmod((%(out_t)s)a, (%(out_t)s)b)"
        return elemwise2(self, '%', other, self, odtype=out_dtype, **kw)

    def __rmod__(self, other):
        out_dtype = get_common_dtype(other, self, True)
        kw = {'broadcast': True}
        if out_dtype.kind == 'f':
            kw['op_tmpl'] = "res = fmod((%(out_t)s)a, (%(out_t)s)b)"
        return elemwise2(other, '%', self, self, odtype=out_dtype, **kw)

    def __imod__(self, other):
        out_dtype = get_common_dtype(self, other, self.dtype == np.float64)
        kw = {'broadcast': True}
        if out_dtype == np.float32:
            kw['op_tmpl'] = "a = fmod((float)a, (float)b)"
        if out_dtype == np.float64:
            kw['op_tmpl'] = "a = fmod((double)a, (double)b)"
        return ielemwise2(self, '%', other, **kw)

    # divmod
    def __divmod__(self, other):
        if not isinstance(other, gpuarray.GpuArray):
            other = np.asarray(other)
        odtype = get_common_dtype(self, other, True)

        a_arg = as_argument(self, 'a', read=True)
        b_arg = as_argument(other, 'b', read=True)
        args = [arg('div', odtype, write=True), arg('mod', odtype, write=True), a_arg, b_arg]

        div = self._empty_like_me(dtype=odtype)
        mod = self._empty_like_me(dtype=odtype)

        if odtype.kind == 'f':
            tmpl = ("div = floor((%(out_t)s)a / (%(out_t)s)b),"
                    "mod = fmod((%(out_t)s)a, (%(out_t)s)b)")
        else:
            tmpl = ("div = (%(out_t)s)a / (%(out_t)s)b,"
                    "mod = a %% b")

        ksrc = tmpl % {'out_t': dtype_to_ctype(odtype)}

        k = GpuElemwise(self.context, ksrc, args)
        k(div, mod, self, other, broadcast=True)
        return (div, mod)

    def __rdivmod__(self, other):
        if not isinstance(other, gpuarray.GpuArray):
            other = np.asarray(other)
        odtype = get_common_dtype(other, self, True)

        a_arg = as_argument(other, 'a', read=True)
        b_arg = as_argument(self, 'b', read=True)
        args = [arg('div', odtype, write=True), arg('mod', odtype, write=True), a_arg, b_arg]

        div = self._empty_like_me(dtype=odtype)
        mod = self._empty_like_me(dtype=odtype)

        if odtype.kind == 'f':
            tmpl = ("div = floor((%(out_t)s)a / (%(out_t)s)b),"
                    "mod = fmod((%(out_t)s)a, (%(out_t)s)b)")
        else:
            tmpl = ("div = (%(out_t)s)a / (%(out_t)s)b,"
                    "mod = a %% b")

        ksrc = tmpl % {'out_t': dtype_to_ctype(odtype)}

        k = GpuElemwise(self.context, ksrc, args)
        k(div, mod, other, self, broadcast=True)
        return (div, mod)

    def __neg__(self):
        return elemwise1(self, '-')

    def __pos__(self):
        return elemwise1(self, '+')

    def __abs__(self):
        if self.dtype.kind == 'u':
            return self.copy()
        if self.dtype.kind == 'f':
            oper = "res = fabs(a)"
        elif self.dtype.itemsize < 4:
            # cuda 5.5 finds the c++ stdlib definition if we don't cast here.
            oper = "res = abs((int)a)"
        else:
            oper = "res = abs(a)"
        return elemwise1(self, None, oper=oper)

    # richcmp
    def __lt__(self, other):
        return compare(self, '<', other, broadcast=True)

    def __le__(self, other):
        return compare(self, '<=', other, broadcast=True)

    def __eq__(self, other):
        return compare(self, '==', other, broadcast=True)

    def __ne__(self, other):
        return compare(self, '!=', other, broadcast=True)

    def __ge__(self, other):
        return compare(self, '>=', other, broadcast=True)

    def __gt__(self, other):
        return compare(self, '>', other, broadcast=True)

    # misc other things
    @property
    def T(self):
        if self.ndim < 2:
            return self
        return self.transpose()

    """
Since these functions are untested (thus probably wrong), we disable them.
    def clip(self, a_min, a_max, out=None):
        oper=('res = a > %(max)s ? %(max)s : '
              '(a < %(min)s ? %(min)s : a)' % dict(min=a_min, max=a_max))
        return elemwise1(self, '', oper=oper, out=out)

    def fill(self, value):
        self[...] = value
"""
    # reductions
    def all(self, axis=None, out=None):
        if self.ndim == 0:
            return self.copy()
        return reduce1(self, '&&', '1', np.dtype('bool'),
                       axis=axis, out=out)

    def any(self, axis=None, out=None):
        if self.ndim == 0:
            return self.copy()
        return reduce1(self, '||', '0', np.dtype('bool'),
                       axis=axis, out=out)

    def prod(self, axis=None, dtype=None, out=None):
        if dtype is None:
            dtype = self.dtype
            # we only upcast integers that are smaller than the plaform default
            if dtype.kind == 'i':
                di = np.dtype('int')
                if di.itemsize > dtype.itemsize:
                    dtype = di
            if dtype.kind == 'u':
                di = np.dtype('uint')
                if di.itemsize > dtype.itemsize:
                    dtype = di
        return reduce1(self, '*', '1', dtype, axis=axis, out=out)

#    def max(self, axis=None, out=None);
#        nd = self.ndim
#        if nd == 0:
#            return self.copy()
#        idx = (0,) * nd
#        n = str(self.__getitem__(idx).__array__())
#        return reduce1(self, '', n, self.dtype, axis=axis, out=out,
#                       oper='max(a, b)')

#    def min(self, axis=None, out=None):
#        nd = self.ndim
#        if nd == 0:
#            return self.copy()
#        idx = (0,) * nd
#        n = str(self.__getitem__(idx).__array__())
#        return reduce1(self, '', n, self.dtype, axis=axis, out=out,
#                       oper='min(a, b)')

    def sum(self, axis=None, dtype=None, out=None):
        if dtype is None:
            dtype = self.dtype
            # we only upcast integers that are smaller than the plaform default
            if dtype.kind == 'i':
                di = np.dtype('int')
                if di.itemsize > dtype.itemsize:
                    dtype = di
            if dtype.kind == 'u':
                di = np.dtype('uint')
                if di.itemsize > dtype.itemsize:
                    dtype = di
        return reduce1(self, '+', '0', dtype, axis=axis, out=out)



import os
import nose.plugins.builtin

from nose.config import Config
from nose.plugins.manager import PluginManager
from numpy.testing.nosetester import NoseTester
from numpy.testing.noseclasses import KnownFailure, NumpyTestProgram


class NoseTester(NoseTester):
    """
    Nose test runner.

    This class enables running nose tests from inside libgpuarray,
    by calling pygpu.test().
    This version is more adapted to what we want than Numpy's one.
    """

    def _test_argv(self, verbose, extra_argv):
        """
        Generate argv for nosetest command

        Parameters
        ----------
        verbose: int
            Verbosity value for test outputs, in the range 1-10.
            Default is 1.
        extra_argv: list
            List with any extra arguments to pass to nosetests.

        """
        # self.package_path = os.path.abspath(self.package_path)
        argv = [__file__, self.package_path]
        argv += ['--verbosity', str(verbose)]
        if extra_argv:
            argv += extra_argv
        return argv

    def _show_system_info(self):
        import pygpu
        # print ("pygpu version %s" % pygpu.__version__)
        pygpu_dir = os.path.dirname(pygpu.__file__)
        print("pygpu is installed in %s" % pygpu_dir)

        super(NoseTester, self)._show_system_info()

    def prepare_test_args(self, verbose=1, extra_argv=None, coverage=False,
                          capture=True, knownfailure=True):
        """
        Prepare arguments for the `test` method.

        Takes the same arguments as `test`.
        """
        # compile argv
        argv = self._test_argv(verbose, extra_argv)

        # numpy way of doing coverage
        if coverage:
            argv += ['--cover-package=%s' % self.package_name,
                     '--with-coverage', '--cover-tests', '--cover-inclusive',
                     '--cover-erase']

        # Capture output only if needed
        if not capture:
            argv += ['-s']

        # construct list of plugins
        plugins = []
        if knownfailure:
            plugins.append(KnownFailure())
        plugins += [p() for p in nose.plugins.builtin.plugins]

        return argv, plugins

    def test(self, verbose=1, extra_argv=None, coverage=False, capture=True,
             knownfailure=True):
        """
        Run tests for module using nose.

        Parameters
        ----------
        verbose: int
            Verbosity value for test outputs, in the range 1-10.
            Default is 1.
        extra_argv: list
            List with any extra arguments to pass to nosetests.
        coverage: bool
            If True, report coverage of pygpu code. Default is False.
        capture: bool
            If True, capture the standard output of the tests, like
            nosetests does in command-line. The output of failing
            tests will be displayed at the end. Default is True.
        knownfailure: bool
            If True, tests raising KnownFailureTest will not be
            considered Errors nor Failure, but reported as "known
            failures" and treated quite like skipped tests.  Default
            is True.

        Returns
        -------
        nose.result.TextTestResult
            The result of running the tests

        """
        # cap verbosity at 3 because nose becomes *very* verbose beyond that
        verbose = min(verbose, 3)
        self._show_system_info()

        cwd = os.getcwd()
        if self.package_path in os.listdir(cwd):
            # The tests give weird errors if the package to test is
            # in current directory.
            raise RuntimeError((
                "This function does not run correctly when, at the time "
                "pygpu was imported, the working directory was pygpu's "
                "parent directory. You should exit your Python prompt, change "
                "directory, then launch Python again, import pygpu, then "
                "launch pygpu.test()."))

        argv, plugins = self.prepare_test_args(verbose, extra_argv, coverage,
                                               capture, knownfailure)

        # The "plugins" keyword of NumpyTestProgram gets ignored if config is
        # specified. Moreover, using "addplugins" instead can lead to strange
        # errors. So, we specify the plugins in the Config as well.
        cfg = Config(includeExe=True, plugins=PluginManager(plugins=plugins))
        t = NumpyTestProgram(argv=argv, exit=False, config=cfg)
        return t.result



import operator
import numpy
from mako.template import Template

from unittest import TestCase
from pygpu import gpuarray, ndgpuarray as elemary
from pygpu.dtypes import dtype_to_ctype, get_common_dtype
from pygpu.elemwise import as_argument, ielemwise2
from pygpu._elemwise import GpuElemwise, arg

from six import PY2

from .support import (guard_devsup, context, gen_gpuarray, check_meta_content)

dtypes_test = ['float32', 'int8', 'uint64']

operators1 = [operator.neg, operator.pos, operator.abs]
operators2 = [operator.add, operator.sub, operator.floordiv,
              operator.mod, operator.mul, operator.truediv,
              operator.eq, operator.ne, operator.lt, operator.le,
              operator.gt, operator.ge]
if PY2:
    operators2.append(operator.div)

ioperators2 = [operator.iadd, operator.isub, operator.ifloordiv,
               operator.imod, operator.imul, operator.itruediv]
if PY2:
    ioperators2.append(operator.idiv)

elems = [2, 0.3, numpy.asarray(3, dtype='int8'),
         numpy.asarray(7, dtype='uint32'),
         numpy.asarray(2.45, dtype='float32')]


def test_elemwise1_ops_array():
    for op in operators1:
        for dtype in dtypes_test:
            yield elemwise1_ops_array, op, dtype


@guard_devsup
def elemwise1_ops_array(op, dtype):
    c, g = gen_gpuarray((50,), dtype, ctx=context, cls=elemary)

    out_c = op(c)
    out_g = op(g)

    assert out_c.shape == out_g.shape
    assert out_c.dtype == out_g.dtype
    assert numpy.allclose(out_c, numpy.asarray(out_g))


def test_elemwise2_ops_array():
    for op in operators2:
        for dtype1 in dtypes_test:
            for dtype2 in dtypes_test:
                yield elemwise2_ops_array, op, dtype1, dtype2, (50,)


def test_ielemwise2_ops_array():
    for op in ioperators2:
        for dtype1 in dtypes_test:
            for dtype2 in dtypes_test:
                yield ielemwise2_ops_array, op, dtype1, dtype2, (50,)


class test_elemwise_output_not_broadcasted(TestCase):
    def test_all(self):
        test_values = [((1, 4), (6, 4)), ((2, 1, 8, 7), (2, 2, 8, 7))]
        for shapea, shapeb in test_values:
            # Sould fail: dimensions are not all equal.
            self.assertRaises(ValueError, self.run_ielemwise2, shapea, shapeb,
                              False)
            # Should fail: broascast should not be done on output.
            self.assertRaises(ValueError, self.run_ielemwise2, shapea, shapeb,
                              True)
            # Should fail: dimensions are not all equal.
            self.assertRaises(ValueError, self.check_elemwise2, shapeb, shapeb,
                              shapea, False)
            # Should fail: broadcast should not be done on output.
            self.assertRaises(ValueError, self.check_elemwise2, shapeb, shapeb,
                              shapea, True)
            # Should pass: output would be done on read-only input.
            self.run_ielemwise2(shapeb, shapea, broadcast=True)
            # Should pass: output would be done on read-only inputs.
            self.check_elemwise2(shapea, shapea, shapeb, broadcast=True)
            self.check_elemwise2(shapea, shapeb, shapeb, broadcast=True)
            self.check_elemwise2(shapeb, shapea, shapeb, broadcast=True)

    @guard_devsup
    def run_ielemwise2(self, shapea, shapeb, broadcast=True):
        na, ga = gen_gpuarray(shapea, ctx=context, cls=elemary)
        nb, gb = gen_gpuarray(shapeb, ctx=context, cls=elemary)
        ielemwise2(ga, '+', gb, broadcast=broadcast)
        na += nb
        assert numpy.allclose(na, numpy.asarray(ga), atol=1e-6)

    @guard_devsup
    def check_elemwise2(self, shapea, shapeb, output_shape, broadcast=True):
        # We rewrite this version of elemwise2 to skip the scaling of output
        # that is done in the official elemwise2 function.
        na, ga = gen_gpuarray(shapea, ctx=context, cls=elemary)
        nb, gb = gen_gpuarray(shapeb, ctx=context, cls=elemary)
        odtype = get_common_dtype(ga, gb, True)
        res = gpuarray.empty(output_shape, dtype=odtype, context=ga.context,
                             cls=ga.__class__)
        a_arg = as_argument(ga, 'a', read=True)
        b_arg = as_argument(gb, 'b', read=True)
        res_arg = as_argument(res, 'res', write=True)
        args = [res_arg, a_arg, b_arg]
        oper = "res = (%(out_t)s)a %(op)s (%(out_t)s)b" % {
            'op': '+', 'out_t': dtype_to_ctype(odtype)}
        k = GpuElemwise(ga.context, oper, args, convert_f16=True)
        k(res, ga, gb, broadcast=broadcast)
        nres = na + nb
        assert numpy.allclose(nres, numpy.asarray(res), atol=1e-6)


@guard_devsup
def elemwise2_ops_array(op, dtype1, dtype2, shape):
    ac, ag = gen_gpuarray(shape, dtype1, ctx=context, cls=elemary)
    bc, bg = gen_gpuarray(shape, dtype2, nozeros=True, ctx=context,
                          cls=elemary)

    out_c = op(ac, bc)
    out_g = op(ag, bg)

    assert out_c.shape == out_g.shape
    assert out_c.dtype == out_g.dtype
    assert numpy.allclose(out_c, numpy.asarray(out_g))


@guard_devsup
def ielemwise2_ops_array(op, dtype1, dtype2, shape):
    incr = 0
    if op == operator.isub and dtype1[0] == 'u':
        # array elements are smaller than 10 by default, so we avoid underflow
        incr = 10
    ac, ag = gen_gpuarray(shape, dtype1, incr=incr, ctx=context,
                          cls=elemary)
    bc, bg = gen_gpuarray(shape, dtype2, nozeros=True, ctx=context,
                          cls=elemary)

    try:
        out_c = op(ac, bc)
    except TypeError:
        # TODO: currently, we use old Numpy semantic and tolerate more case.
        # So we can't test that we raise the same error
        return
    out_g = op(ag, bg)

    assert out_g is ag
    assert numpy.allclose(out_c, numpy.asarray(out_g), atol=1e-6)


def test_elemwise_f16():
    yield elemwise1_ops_array, operator.neg, 'float16'
    yield elemwise2_ops_array, operator.add, 'float16', 'float16', (50,)
    yield ielemwise2_ops_array, operator.iadd, 'float16', 'float16', (50,)


def test_elemwise2_ops_mixed():
    for op in operators2:
        for dtype in dtypes_test:
            for elem in elems:
                yield elemwise2_ops_mixed, op, dtype, (50,), elem


def test_ielemwise2_ops_mixed():
    for op in ioperators2:
        for dtype in dtypes_test:
            for elem in elems:
                yield ielemwise2_ops_mixed, op, dtype, (50,), elem


@guard_devsup
def elemwise2_ops_mixed(op, dtype, shape, elem):
    c, g = gen_gpuarray(shape, dtype, ctx=context, cls=elemary)

    out_c = op(c, elem)
    out_g = op(g, elem)

    assert out_c.shape == out_g.shape
    assert out_c.dtype == out_g.dtype
    assert numpy.allclose(out_c, numpy.asarray(out_g))

    c, g = gen_gpuarray(shape, dtype, nozeros=True, ctx=context,
                        cls=elemary)
    out_c = op(elem, c)
    out_g = op(elem, g)

    assert out_c.shape == out_g.shape
    assert out_c.dtype == out_g.dtype
    assert numpy.allclose(out_c, numpy.asarray(out_g))


@guard_devsup
def ielemwise2_ops_mixed(op, dtype, shape, elem):
    incr = 0
    if op == operator.isub and dtype[0] == 'u':
        # array elements are smaller than 10 by default, so we avoid underflow
        incr = 10
    c, g = gen_gpuarray(shape, dtype, incr=incr, ctx=context,
                        cls=elemary)

    try:
        out_c = op(c, elem)
    except TypeError:
        # TODO: currently, we use old Numpy semantic and tolerate more case.
        # So we can't test that we raise the same error
        return
    out_g = op(g, elem)

    assert out_g is g
    assert out_c.shape == out_g.shape
    assert out_c.dtype == out_g.dtype
    assert numpy.allclose(out_c, numpy.asarray(out_g))


def test_divmod():
    for dtype1 in dtypes_test:
        for dtype2 in dtypes_test:
            yield divmod_array, dtype1, dtype2, (50,)
    for dtype in dtypes_test:
        for elem in elems:
            yield divmod_mixed, dtype, (50,), elem


@guard_devsup
def divmod_array(dtype1, dtype2, shape):
    ac, ag = gen_gpuarray(shape, dtype1, ctx=context, cls=elemary)
    bc, bg = gen_gpuarray(shape, dtype2, nozeros=True, ctx=context,
                          cls=elemary)

    out_c = divmod(ac, bc)
    out_g = divmod(ag, bg)

    assert out_c[0].shape == out_g[0].shape
    assert out_c[1].shape == out_g[1].shape
    assert out_c[0].dtype == out_g[0].dtype
    assert out_c[1].dtype == out_g[1].dtype
    assert numpy.allclose(out_c[0], numpy.asarray(out_g[0]))
    assert numpy.allclose(out_c[1], numpy.asarray(out_g[1]))


@guard_devsup
def divmod_mixed(dtype, shape, elem):
    c, g = gen_gpuarray(shape, dtype, nozeros=True, ctx=context,
                        cls=elemary)

    out_c = divmod(c, elem)
    out_g = divmod(g, elem)

    assert out_c[0].shape == out_g[0].shape
    assert out_c[1].shape == out_g[1].shape
    assert out_c[0].dtype == out_g[0].dtype
    assert out_c[1].dtype == out_g[1].dtype
    assert numpy.allclose(out_c[0], numpy.asarray(out_g[0]))
    assert numpy.allclose(out_c[1], numpy.asarray(out_g[1]))

    out_c = divmod(elem, c)
    out_g = divmod(elem, g)

    assert out_c[0].shape == out_g[0].shape
    assert out_c[1].shape == out_g[1].shape
    assert out_c[0].dtype == out_g[0].dtype
    assert out_c[1].dtype == out_g[1].dtype
    assert numpy.allclose(out_c[0], numpy.asarray(out_g[0]))
    assert numpy.allclose(out_c[1], numpy.asarray(out_g[1]))


def test_elemwise_bool():
    a = gpuarray.empty((2,), context=context)
    exc = None
    try:
        bool(a)
    except ValueError as e:
        exc = e
    assert exc is not None
    a = gpuarray.zeros((1,), context=context)
    assert not bool(a)
    a = gpuarray.zeros((), context=context)
    assert not bool(a)


def test_broadcast():
    for shapea, shapeb in [((3, 5), (3, 5)),
                           ((1, 5), (3, 5)),
                           ((3, 5), (3, 1)),
                           ((1, 5), (3, 1)),
                           ((3, 1), (3, 5)),
                           ((3, 5), (3, 1)),
                           ((1, 1), (1, 1)),
                           ((3, 4, 5), (4, 5)),
                           ((4, 5), (3, 4, 5)),
                           ((), ())]:
        yield broadcast, shapea, shapeb


def broadcast(shapea, shapeb):
    ac, ag = gen_gpuarray(shapea, 'float32', ctx=context, cls=elemary)
    bc, bg = gen_gpuarray(shapeb, 'float32', ctx=context, cls=elemary)

    rc = ac + bc
    rg = ag + bg

    check_meta_content(rg, rc)


_inf_preamb_tpl = Template('''
WITHIN_KERNEL ${flt}
infinity() {return INFINITY;}

WITHIN_KERNEL ${flt}
neg_infinity() {return -INFINITY;}
''')


def test_infinity():
    for dtype in ['float32', 'float64']:
        ac, ag = gen_gpuarray((2,), dtype, ctx=context, cls=elemary)
        out_g = ag._empty_like_me()
        flt = 'ga_float' if dtype == 'float32' else 'ga_double'
        out_arg = arg('out', out_g.dtype, scalar=False, read=False, write=True)
        preamble = _inf_preamb_tpl.render(flt=flt)

        # +infinity
        ac[:] = numpy.inf
        expr_inf = 'out = infinity()'
        kernel = GpuElemwise(context, expr_inf, [out_arg],
                             preamble=preamble)
        kernel(out_g)
        assert numpy.array_equal(ac, numpy.asarray(out_g))

        # -infinity
        ac[:] = -numpy.inf
        expr_neginf = 'out = neg_infinity()'
        kernel = GpuElemwise(context, expr_neginf, [out_arg],
                             preamble=preamble)
        kernel(out_g)
        assert numpy.array_equal(ac, numpy.asarray(out_g))
